==> gpsshogi_0.3.0.orig.tar.gz: gpsshogi_0.3.0.orig/sample/gpsshell/book.cc <==
#include "book.h"
#include "osl/record/kanjiPrint.h"
#include "osl/record/csa.h"
#ifndef MINIMAL_GPSSHELL
#  include <gsl_cdf.h>
#endif
#include <boost/shared_ptr.hpp>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <cmath>
#include <iostream>

using namespace boost::lambda;

int gpsshell::
Book::getStateIndex(const osl::SimpleState& state, 
                    const osl::vector<osl::Move>& moves) const
{
  int state_index = the_book->getStateIndex(moves);
  if (state_index < 0)
    state_index = the_book->getStateIndex(state); // it could take a long time
  return state_index;
}

void gpsshell::
Book::showState(const osl::SimpleState& state, 
                const osl::vector<osl::Move>& moves,
                int max_count) const
{
  osl::record::KanjiPrint printer(std::cerr, 
                                  boost::shared_ptr<osl::record::Characters>(
                                            new osl::record::KIFCharacters())
                                  );

  const int state_index = getStateIndex(state, moves);
  if (state_index < 0)
  {
    std::cout << "The current position not found in the book" << std::endl;
    return;
  }

  const osl::Player turn   = state.getTurn();
  const int black_win = the_book->getBlackWinCount(state_index);
  const int white_win = the_book->getWhiteWinCount(state_index);
  const int trial     = black_win + white_win;
  const int nwins     = (turn == osl::BLACK ? black_win : white_win); 
  double win_rate     = 0.0;
  double deviation    = 0.0;
  double cd           = 0.0; // cumulative distribution
  if (trial > 0)
  {
    win_rate = 1.0 * nwins / trial;
#ifndef MINIMAL_GPSSHELL
    cd = gsl_cdf_binomial_P(nwins, 0.5, trial);
#else
    cd = 0;
#endif
    deviation = sqrt( win_rate * (1-win_rate) / trial);
  }

  std::cout << boost::format("State_index: % 7d\n")   % state_index
            << boost::format("Black win:   % 7d\n")   % black_win
            << boost::format("White win:   % 7d\n")   % white_win
            <<               "Binomial:\n" 
            << boost::format("  E(X) [win rate] : % 7.4f\n") % win_rate
            << boost::format("  D(X):             % 7.4f\n") % deviation
            << boost::format("  cumulative:       % 7.4f\n") % cd;

  std::cout << "\n";
  printer.print(state);
  std::cout << "\n";

  gpsshell::WMoveStatsContainer wstats = getWMoveStats(state, moves);
  std::cout << boost::format("found %d moves\n") % wstats.size();

  BOOST_FOREACH(const gpsshell::WMoveStats& st, wstats)
  {
    if (--max_count < 0) break;

    std::cout << boost::format("[% 6d] %s  win:% 7d loss:% 7d  rate:% 7.4f - % 7.4f = % 7.4f\n") 
                                % st.getWmove().getWeight() 
                                % osl::record::csa::show(st.getWmove().getMove())
                                % st.nwins()
                                % st.nlosses()
                                % st.winRate()
                                % st.deviation()
                                % (st.winRate() - st.deviation());
  }
}

gpsshell::WMoveContainer gpsshell::
Book::getMoves(const osl::SimpleState& state, 
               const osl::vector<osl::Move>& moves) const
{
  const gpsshell::WMoveStatsContainer wstats
    = getWMoveStats(state, moves);
  
  gpsshell::WMoveContainer wmoves;

  if (wstats.empty())
    return wmoves; // empty
  
  BOOST_FOREACH(const gpsshell::WMoveStats& ws, wstats)
  {
    wmoves.push_back(ws.getWmove());
  }

  return wmoves;
}

gpsshell::WMoveStatsContainer gpsshell::
Book::getWMoveStats(const osl::SimpleState& state, 
                    const osl::vector<osl::Move>& moves) const
{
  const int state_index = getStateIndex(state, moves);
  gpsshell::WMoveStatsContainer wstats;

  if (state_index < 0)
    return wstats; // empty

  {
    WMoveContainer wmoves = the_book->getMoves(state_index);
    wstats.reserve(wmoves.size());
    BOOST_FOREACH(const osl::record::opening::WMove& wmove, wmoves)
    {
      const int next_state_index = wmove.getStateIndex();
      const int black_win        = the_book->getBlackWinCount(next_state_index);
      const int white_win        = the_book->getWhiteWinCount(next_state_index);
      gpsshell::WMoveStats st(wmove, black_win, white_win);
      wstats.push_back(st);
    }
  }

  /*
   * Sort:
   *   weight such as wegiht > 0
   *   #wins such as weight == 0
   */
  std::sort(wstats.begin(), wstats.end(), gpsshell::WMoveStatsSortByWeight());
  gpsshell::WMoveStatsContainer::iterator it = wstats.begin();
  for (/* none */; it != wstats.end(); ++it)
  {
    if (it->getWmove().getWeight() < 1)
      break;
  }
  std::sort(it, wstats.end(), gpsshell::WMoveStatsSortByWins());
  return wstats;
}

// ==============================================
// Utility functions
// ==============================================

int gpsshell::getMaxWeight(const gpsshell::WMoveContainer& moves)
{
  int max_weight = 0;
  BOOST_FOREACH(const osl::record::opening::WMove& wmove, moves) {
    if (max_weight < wmove.getWeight()) {
      max_weight = wmove.getWeight();
    }
  }
  return max_weight;
}

int gpsshell::getSumOfWeights(const gpsshell::WMoveContainer& moves)
{
  int sum = 0;
  BOOST_FOREACH(const osl::record::opening::WMove& wmove, moves) {
    sum += wmove.getWeight();
  }
  return sum;
}

struct WeightLessThan
{
  int criteria;

  explicit WeightLessThan(const int criteria)
    : criteria(criteria)
  {}

  bool operator()(const osl::record::opening::WMove& wmove) const
  {
    return (wmove.getWeight() < criteria ? true : false);
  }
};

void gpsshell::deleteLessWeightedMoves(gpsshell::WMoveContainer& wmoves,
                                            const int criteria)
{
  wmoves.erase(std::remove_if(wmoves.begin(), wmoves.end(), WeightLessThan(criteria)), 
               wmoves.end());
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:

==> gpsshogi_0.4.2.orig.tar.gz: gpsshogi-0.4.2.orig/sample/gpsshell/book.cc <==
#include "book.h"
#include "osl/record/kanjiPrint.h"
#include "osl/record/csa.h"
#ifndef MINIMAL_GPSSHELL
#  include <gsl_cdf.h>
#endif
#include <boost/shared_ptr.hpp>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <cmath>
#include <iostream>

using namespace boost::lambda;

int gpsshell::
Book::getStateIndex(const osl::SimpleState& state, 
                    const osl::vector<osl::Move>& moves) const
{
  int state_index = the_book->getStateIndex(moves);
  if (state_index < 0)
    state_index = the_book->getStateIndex(state); // it could take a long time
  return state_index;
}

void gpsshell::
Book::showState(const osl::SimpleState& state, 
                const osl::vector<osl::Move>& moves,
                int max_count) const
{
  osl::record::KanjiPrint printer(std::cerr, 
                                  boost::shared_ptr<osl::record::Characters>(
                                            new osl::record::KIFCharacters())
                                  );

  const int state_index = getStateIndex(state, moves);
  if (state_index < 0)
  {
    std::cout << "The current position not found in the book" << std::endl;
    return;
  }

  const osl::Player turn   = state.turn();
  const int black_win = the_book->getBlackWinCount(state_index);
  const int white_win = the_book->getWhiteWinCount(state_index);
  const int trial     = black_win + white_win;
  const int nwins     = (turn == osl::BLACK ? black_win : white_win); 
  double win_rate     = 0.0;
  double deviation    = 0.0;
  double cd           = 0.0; // cumulative distribution
  if (trial > 0)
  {
    win_rate = 1.0 * nwins / trial;
#ifndef MINIMAL_GPSSHELL
    cd = gsl_cdf_binomial_P(nwins, 0.5, trial);
#else
    cd = 0;
#endif
    deviation = sqrt( win_rate * (1-win_rate) / trial);
  }

  std::cout << boost::format("State_index: % 7d\n")   % state_index
            << boost::format("Black win:   % 7d\n")   % black_win
            << boost::format("White win:   % 7d\n")   % white_win
            <<               "Binomial:\n" 
            << boost::format("  E(X) [win rate] : % 7.4f\n") % win_rate
            << boost::format("  D(X):             % 7.4f\n") % deviation
            << boost::format("  cumulative:       % 7.4f\n") % cd;

  std::cout << "\n";
  printer.print(state);
  std::cout << "\n";

  gpsshell::WMoveStatsContainer wstats = getWMoveStats(state, moves);
  std::cout << boost::format("found %d moves\n") % wstats.size();

  BOOST_FOREACH(const gpsshell::WMoveStats& st, wstats)
  {
    if (--max_count < 0) break;

    std::cout << boost::format("[% 6d] %s  win:% 7d loss:% 7d  rate:% 7.4f - % 7.4f = % 7.4f\n") 
                                % st.getWmove().getWeight() 
                                % osl::record::csa::show(st.getWmove().getMove())
                                % st.nwins()
                                % st.nlosses()
                                % st.winRate()
                                % st.deviation()
                                % (st.winRate() - st.deviation());
  }
}

gpsshell::WMoveContainer gpsshell::
Book::getMoves(const osl::SimpleState& state, 
               const osl::vector<osl::Move>& moves) const
{
  const gpsshell::WMoveStatsContainer wstats
    = getWMoveStats(state, moves);
  
  gpsshell::WMoveContainer wmoves;

  if (wstats.empty())
    return wmoves; // empty
  
  BOOST_FOREACH(const gpsshell::WMoveStats& ws, wstats)
  {
    wmoves.push_back(ws.getWmove());
  }

  return wmoves;
}

gpsshell::WMoveStatsContainer gpsshell::
Book::getWMoveStats(const osl::SimpleState& state, 
                    const osl::vector<osl::Move>& moves) const
{
  const int state_index = getStateIndex(state, moves);
  gpsshell::WMoveStatsContainer wstats;

  if (state_index < 0)
    return wstats; // empty

  {
    WMoveContainer wmoves = the_book->getMoves(state_index);
    wstats.reserve(wmoves.size());
    BOOST_FOREACH(const osl::record::opening::WMove& wmove, wmoves)
    {
      const int next_state_index = wmove.getStateIndex();
      const int black_win        = the_book->getBlackWinCount(next_state_index);
      const int white_win        = the_book->getWhiteWinCount(next_state_index);
      gpsshell::WMoveStats st(wmove, black_win, white_win);
      wstats.push_back(st);
    }
  }

  /*
   * Sort:
   *   weight such as wegiht > 0
   *   #wins such as weight == 0
   */
  std::sort(wstats.begin(), wstats.end(), gpsshell::WMoveStatsSortByWeight());
  gpsshell::WMoveStatsContainer::iterator it = wstats.begin();
  for (/* none */; it != wstats.end(); ++it)
  {
    if (it->getWmove().getWeight() < 1)
      break;
  }
  std::sort(it, wstats.end(), gpsshell::WMoveStatsSortByWins());
  return wstats;
}

// ==============================================
// Utility functions
// ==============================================

int gpsshell::getMaxWeight(const gpsshell::WMoveContainer& moves)
{
  int max_weight = 0;
  BOOST_FOREACH(const osl::record::opening::WMove& wmove, moves) {
    if (max_weight < wmove.getWeight()) {
      max_weight = wmove.getWeight();
    }
  }
  return max_weight;
}

int gpsshell::getSumOfWeights(const gpsshell::WMoveContainer& moves)
{
  int sum = 0;
  BOOST_FOREACH(const osl::record::opening::WMove& wmove, moves) {
    sum += wmove.getWeight();
  }
  return sum;
}

struct WeightLessThan
{
  int criteria;

  explicit WeightLessThan(const int criteria)
    : criteria(criteria)
  {}

  bool operator()(const osl::record::opening::WMove& wmove) const
  {
    return (wmove.getWeight() < criteria ? true : false);
  }
};

void gpsshell::deleteLessWeightedMoves(gpsshell::WMoveContainer& wmoves,
                                            const int criteria)
{
  wmoves.erase(std::remove_if(wmoves.begin(), wmoves.end(), WeightLessThan(criteria)), 
               wmoves.end());
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
